/*=======================================================================

  Program:  Distance Transforms
  Module:   $Id: compare3.c,v 0.1 $
  Date:     $Date: 2003/05/26 10:00:11 $
  Language: C
  
  Author:   Olivier Cuisenaire, http://ltswww.epfl.ch/~cuisenai

  Changed by Francisco de Assis Zampirolli, http://www.dca.fee.unicamp.br/~fz
  
  Usage:    compare3 <size> 


  Description: 

   Test program that compares the result of the various DT algorithms 
   to that of Saito's. Checks which algorithms compute exact EDTs and 
   which don't. It also displays the needed CPU times to allow complexity
   comparisons. 

   There is one optional argument which is the size of the square image
   that is used for testing. The test image is a circle centered on the
   image from which we compute distances both inside and outside. 

  Reference for the implementation:

   "Distance transformations: fast algorithms and applications 
   to medical image processing", Olivier Cuisenaire's Ph.D. Thesis, October 
   1999, Universit catholique de Louvain, Belgium.
    
  Terms of Use: 

   You can use/modify this program for any use you wish, provided you cite 
   the above references in any publication about it. 
    
  Disclaimer:

   In no event shall the authors or distributors be liable to any party for 
   direct, indirect, special, incidental, or consequential damages arising out 
   of the use of this software, its documentation, or any derivatives thereof, 
   even if the authors have been advised of the possibility of such damage.
   
   The authors and distributors specifically disclaim any warranties, including,
   but not limited to, the implied warranties of merchantability, fitness for a
   particular purpose, and non-infringement.  this software is provided on an 
   "as is" basis, and the authors and distributors have no obligation to provide
   maintenance, support, updates, enhancements, or modifications.
   
   =========================================================================*/

#include "distmaps.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

void parse(int argc, char * argv[],int size[2]);
void draw(int *im, int size[2]);
int diff(int *im1, int *im2, int size[2]);

int main(int argc, char * argv[])
{
  int *image1, *image2, *image3;
  int size[2];
  clock_t t; float time;

  parse(argc,argv,size);

  printf("Euclidean distance transformation algorithms comparison. \n\nUsage: compare [N] \n\nThis program compares the speed of various EDT algorithms \nfor a square test image of size NxN. The test image consists\nof a pixel centered on the image. Default image size is 1024. \n\nImage size for this test: %d x %d \n\n",size[0],size[1]);
  
  image1 = (int*) calloc ( size[0]*size[1], sizeof(int));
  image2 = (int*) calloc ( size[0]*size[1], sizeof(int));
  image3 = (int*) calloc ( size[0]*size[1], sizeof(int));

  printf("Computing Saito DT .... "); fflush(stdout);
  draw(image1,size);
  t = clock();
  distmap_saito2d(image1,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms\n",time*1000);
 
  printf("Computing Saito XY .... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_saito2d_xy(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Saito YX .... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_saito2d_yx(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Eggers ...... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_eggers(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Maurer ...... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_maurer(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing 4SSED ....... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_4ssed(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing 4SSED+ ...... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_4ssedp(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing PSN ......... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_psn(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing PMN ......... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_pmn(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing PMON ........ "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_pmon(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Meijster .... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  fz_edt_meijster(image2,image3,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image3,size));

  printf("Computing LZ .......... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  lz_edt2(image2,image3,size[0],size[1],1);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image3,size));

  printf("Computing EroDir ...... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_erosion_dir(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing EroPro ...... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_erosion_pro(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing EroPar ...... "); fflush(stdout);
  draw(image2,size);
  t = clock();
  distmap_erosion_par(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  free(image1); free(image2); free(image3);

  return 0;
}

void parse(int argc, char * argv[],int size[2])
{
	int val;
	
	size[0]=size[1]=1024;
	
	if( argc >= 2 ) { 
		val = atoi(argv[1]);
		if( val>10 && val < 10000 ) 
			size[0]=size[1]=val;
	}
}

void draw(int *im, int size[2])
{
	int i,j;
	int maxval;
	float rx, ry;
	
	rx = (float)( size[0] * 0.5 ) ;
	ry = (float)( size[1] * 0.5 ) ;
	maxval = size[0]*size[0]+size[1]*size[1]+1;
	
	for(j=0;j<size[1];j++)
		for(i=0;i<size[0];i++)
			im[i+j*size[0]] = maxval;
	im[((int)rx)+((int) ry)*size[0]] = 0;
}

int diff(int *im1, int *im2, int size[2])
{
	int i,j,pt;
	int count=0;
	
	pt=0;
	for(j=0;j<size[1];j++)
		for(i=0;i<size[0];i++,pt++)
			if( im1[pt] != im2[pt] ) 
				count++;
			
			return(count);
}
