/*=======================================================================

  Program:  Distance Transforms
  Module:   $Id: compare5.c,v 0.1$
  Date:     $Date: 2003/05/26 10:00:11 $
  Language: C
  
  Author:   Olivier Cuisenaire, http://ltswww.epfl.ch/~cuisenai

  Changed by Francisco de Assis Zampirolli, http://www.dca.fee.unicamp.br/~fz
  
  Usage:    compare5 <size> 

  Description: 

   Test program that compares the result of the various DT algorithms 
   to that of Saito's. Checks which algorithms compute exact EDTs and 
   which don't. It also displays the needed CPU times to allow complexity
   comparisons. 

   There is one optional argument which is the size of the square image
   that is used for testing. The test image is a circle centered on the
   image from which we compute distances both inside and outside. 

  Reference for the implementation:

   "Distance transformations: fast algorithms and applications 
   to medical image processing", Olivier Cuisenaire's Ph.D. Thesis, October 
   1999, Universit catholique de Louvain, Belgium.
    
  Terms of Use: 

   You can use/modify this program for any use you wish, provided you cite 
   the above references in any publication about it. 
    
  Disclaimer:

   In no event shall the authors or distributors be liable to any party for 
   direct, indirect, special, incidental, or consequential damages arising out 
   of the use of this software, its documentation, or any derivatives thereof, 
   even if the authors have been advised of the possibility of such damage.
   
   The authors and distributors specifically disclaim any warranties, including,
   but not limited to, the implied warranties of merchantability, fitness for a
   particular purpose, and non-infringement.  this software is provided on an 
   "as is" basis, and the authors and distributors have no obligation to provide
   maintenance, support, updates, enhancements, or modifications.
   
   =========================================================================*/

#include "distmaps.h"

int diff(int *im1, int *im2, int size[2]);
int *readPGM(int size[2], char *matrixFileName, int t);
void printMatrix(int *image, int size[2]);

int main(int argc, char * argv[])
{
  int *imageIn, *image1, *image2, *image3;
  int size[2];
  clock_t t; float time;
 
  if (argc != 3) {
    printf("Use:\n\tcompare6 imagemIn.pgm vezes\n") ;
    return 0;
  }

  imageIn = readPGM(size, argv[1], atoi(argv[2]));
  image1 = (int *)malloc(size[0]*size[1]*sizeof(int)) ;
  image2 = (int *)malloc(size[0]*size[1]*sizeof(int)) ;
  image3 = (int *)malloc(size[0]*size[1]*sizeof(int)) ;

  printf("Euclidean distance transformation algorithms comparison. \n\nUsage: compare [N] \n\nThis program compares the speed of various EDT algorithms \nfor a square test image of size NxN. \n\nImage size for this test: %d x %d \n\n",size[0],size[1]);
  
  printf("Computing Saito DT .... "); fflush(stdout);
  memcpy(image1,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_saito2d(image1,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms\n",time*1000);

  printf("Computing Saito XY .... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_saito2d_xy(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Saito YX .... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_saito2d_yx(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Eggers ...... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_eggers(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Maurer ...... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_maurer(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing 4SSED ....... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_4ssed(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing 4SSED+ ...... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_4ssedp(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing PSN ......... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_psn(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing PMN ......... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_pmn(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing PMON ........ "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_pmon(image2,size);  
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing Meijster .... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  fz_edt_meijster(image2,image3,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image3,size));

  printf("Computing LZ .......... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  lz_edt2(image2,image3,size[0],size[1],1);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image3,size));

  printf("Computing EroDir ...... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_erosion_dir(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing EroPro ...... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_erosion_pro(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  printf("Computing EroPar ...... "); fflush(stdout);
  memcpy(image2,imageIn,size[0]*size[1]*sizeof(int));
  t = clock();
  distmap_erosion_par(image2,size);
  time = (float)( clock() - t ) / CLOCKS_PER_SEC;
  printf("%10.1f ms ",time*1000);
  printf("... Diffs: %d \n",diff(image1,image2,size));

  free(image1); free(image2); free(image3); free(imageIn);

  return 0;
}

int diff(int *im1, int *im2, int size[2])
{
	int i,j,pt;
	int count=0;
	
	pt=0;
	for(j=0;j<size[1];j++)
		for(i=0;i<size[0];i++,pt++)
			if( im1[pt] != im2[pt] ) 
				count++;
			
			return(count);
}


/* l imagem retornando imagem {0,maxval} de tamanho t vezes imagem de entrada */
int *readPGM(int size[2], char *matrixFileName, int t) {
	int i, j, *A, *B ;
	char type[2];
	int max, maxval;
	FILE *arq ;
	int w,h;
	
	if((arq = fopen(matrixFileName, "r"))==NULL) {
		printf("O arquivo nao pode ser aberto\n") ;
		exit(1) ;
	}
	
	fscanf(arq,"%s",type); 	   /* l tipo */
	fscanf(arq,"%d %d",&w,&h); /* l dimenses */
	fscanf(arq,"%d",&max); 	   /* l maior nvel de cinza */

	A = (int *)malloc(w*h*sizeof(int)) ;
	
	for (i=0; i < w; i++) for (j=0; j < h; j++)	fscanf(arq,"%d",(A+i*w+j)) ;
		
	fclose(arq);
		
	size[0]=t*w; 	size[1]=t*h;
	B = (int *)malloc(size[0]*size[1]*sizeof(int)) ;

	for (i=0; i < size[0]; i++)	for (j=0; j < size[1]; j++)	B[i*size[0]+j]=A[(i%w)*w+(j%h)];

	free(A);

	maxval = size[0]*size[0]+size[1]*size[1]+1;
	for (i=0; i < size[0]*size[1]; i++)	if (B[i]) B[i]=maxval;

	return (B);	
}

void printMatrix(int *image, int size[2]) {
	int i;
    for (i=0;i<size[0]*size[1];i++) {
		if (!(i%size[0])) printf("\n");
		printf("%3d ",image[i]);
	}
}
