/*=======================================================================

  Program:  Distance Transforms
  Module:   $Id: eggers.c,v 1.2 2002/04/24 15:39:43 cuisenai Exp $
  Date:     $Date: 2002/04/24 15:39:43 $
  Language: C
  
  Author:   Olivier Cuisenaire, http://ltswww.epfl.ch/~cuisenai
            
  Descr.: distmap_eggers(int *map, int n[2]) 

   Takes an image (map) of integers of size n[0]*n[1] as input. It 
   computes for all pixels the square of the Euclidean distance to 
   the nearest zero pixel using Eggers' algorithm. 

  Reference for the algorithm: 
  
   H. Eggers, "Two fast Euclidean distance transformations in Z2 based on 
   sufficient propagation", Computer Vision and Image Understanding, 69(1),
   pp. 106-116, 1998

  Reference for the implementation:

   Chapter 2 of "Distance transformations: fast algorithms and applications 
   to medical image processing", Olivier Cuisenaire's Ph.D. Thesis, October 
   1999, Universit catholique de Louvain, Belgium.
    
  Terms of Use: 

   You can use/modify this program for any use you wish, provided you cite 
   the above references in any publication that uses it. 
    
  Disclaimer:

   In no event shall the authors or distributors be liable to any party for 
   direct, indirect, special, incidental, or consequential damages arising out 
   of the use of this software, its documentation, or any derivatives thereof, 
   even if the authors have been advised of the possibility of such damage.
   
   The authors and distributors specifically disclaim any warranties, including,
   but not limited to, the implied warranties of merchantability, fitness for a
   particular purpose, and non-infringement.  this software is provided on an 
   "as is" basis, and the authors and distributors have no obligation to provide
   maintenance, support, updates, enhancements, or modifications.
   
=========================================================================*/

#include "distmaps.h"

struct list *create_list(int size){
	
	struct list *list;
	
	list=(struct list *)malloc(sizeof(struct list));
	list->start=0; list->end=0;
	list->pixels=(struct pixel*)calloc(size,sizeof(struct pixel));
	list->size=size;
	
	return(list);
}

void update_list(int size, struct list *list){
	list->size = size;
	list->pixels=(struct pixel*)realloc(list->pixels,sizeof(struct pixel)*size);
}

void delete_list(struct list *list){
	free(list->pixels);
	free(list);
}

void delete_LIST(struct LIST *list){	
	free(list->pixels);
	free(list);
}

struct LIST *create_LIST(int size){	
	struct LIST *list;
	
	list=(struct LIST *)malloc(sizeof(struct LIST));
	list->start=0; list->end=0;
	list->pixels=(struct PIXEL*)calloc(size,sizeof(struct PIXEL));
	list->size=size;
	
	return(list);
}

void update_LIST(int size, struct LIST *list){
	list->size = size;
	list->pixels=(struct PIXEL*)realloc(list->pixels,sizeof(struct PIXEL)*size);
}

void empty_list(struct list *list){
	list->start=0;
	list->end=0;
}

void empty_LIST(struct LIST *LIST){
	LIST->start=0;
	LIST->end=0;
}

void add_to_LIST(struct LIST *LIST, short x, short y, char k){
	struct PIXEL *Z;
	
	if ( LIST->end >= LIST->size-2 ) 
		update_LIST(LIST->size*2,LIST);
	Z=LIST->pixels+LIST->end;
	Z->x=x; Z->y=y; Z->k=k;
	LIST->end++;
	
}

void add_to_list(struct list *list, short x, short y, char k, int d){
	struct pixel *z;
	
	if ( list->end >= list->size-2 ) 
		update_list(list->size*2,list);
	z=list->pixels+list->end;
	z->x=x; z->y=y; z->k=k; z->d=d;
	list->end++;
}

int pop_list(struct list *list,struct pixel *z){
	int r;
	if(list->end==list->start) r=0;
	else
    {
		*z=list->pixels[list->start];
		list->start++;
		r=1;
    }	 
	return(r);    
}

int pop_LIST(struct LIST *LIST, struct PIXEL *Z){
	int r;
	if(LIST->end==LIST->start) r=0;
	else
    {
		*Z=LIST->pixels[LIST->start];
		LIST->start++;
		r=1;
    }	 
	return(r);
}

int inrange(short x, short y){
	extern short sx,sy;
	int r=1;
	
	if(x<0) r=0;
	if(x>=sx) r=0;
	if(y<0) r=0;
	if(y>=sy) r=0;
	return(r);
}

int distmap_leggers(int *map, int size[2]);

int distmap_eggers(int *map, int n[2])
{

  extern struct LIST *LIST1, *LIST2; 
  extern struct list *list1, *list2;
  
  LIST1=create_LIST(LIST_SIZE);
  LIST2=create_LIST(LIST_SIZE);
  list1=create_list(LIST_SIZE);
  list2=create_list(LIST_SIZE);
  
  distmap_leggers(map,n);

  delete_list(list1);
  delete_list(list2);
  delete_LIST(LIST1);
  delete_LIST(LIST2);

  return(1);
}  

int distmap_leggers(int *map, int size[2])
{
  extern short sx,sy;

  short dx[8] = {1,1,0,-1,-1,-1,0,1} ;
  short dy[8] = {0,-1,-1,-1,0,1,1,1} ;
  
  short x,y;
  char k;
  int d;
  int *pt;
  
  int iter, DIRECT, INDIRECT, adddir, addind;
  int MAXD, non_empty;

  extern struct LIST *LIST1, *LIST2; 
  struct LIST *TLIST;
  extern struct list *list1, *list2;
  struct list *tlist;
  struct pixel z;
  struct PIXEL Z;

  /* memory allocation */


  /* initialisation */
  
  sx=size[0]; sy=size[1];
  iter=0;
  INDIRECT=0;
  MAXD=sx*sx+sy*sy;

  pt=(int*)map; 
  for(y=0;y<sy;y++)
    for(x=0;x<sx;x++,pt++)
      {
	if(*pt==0) 
	  {  
	    if(x<sx-1)
	      if(*(pt+1)!=0)
		add_to_LIST(LIST1,x,y,1);
	    if(y>0)
	      if(*(pt-sx)!=0)
		add_to_LIST(LIST1,x,y,3);
	    if(x>0)
	      if(*(pt-1)!=0)
		add_to_LIST(LIST1,x,y,5);
	    if(y<sy-1)
	      if(*(pt+sx)!=0)
		add_to_LIST(LIST1,x,y,7);
	  }
	else
	  *pt=MAXD;
      }
      


  /* Iterations */

  non_empty=1;
  while(non_empty) {  



    non_empty=0;
    iter++;
    adddir=2*iter-1; addind=2*adddir;
    DIRECT=INDIRECT+adddir;
    INDIRECT=INDIRECT+addind;

    while(pop_list(list1,&z)) {
      non_empty=1;      
      /* propagate(z,k,d+adddir) */
      k=z.k;
      x=z.x+dx[k];
      y=z.y+dy[k];
      d=z.d+adddir;
      pt=map+x+sx*y;
      if(inrange(x,y)) if(*pt>d)
	{
	  *pt=d;
	  add_to_list(list2,x,y,k,d);
	}
      /* end of propagate(z,k,d+adddir) */
    }


    while(pop_LIST(LIST1,&Z)) {
      non_empty=1;
      /* propagate(z,k+1,DIRECT) */
      k=(Z.k+1)%8;
      x=Z.x+dx[k];
      y=Z.y+dy[k];
      pt=map+x+sx*y;
      if(inrange(x,y)) if(*pt>DIRECT)
	{
	  *pt=DIRECT;
	  add_to_list(list2,x,y,k,DIRECT);
	}
      /* propagate(z,k-1,DIRECT) */
      k=(Z.k+7)%8;
      x=Z.x+dx[k];
      y=Z.y+dy[k];
      pt=map+x+sx*y;
      if(inrange(x,y)) if(*pt>DIRECT)
	{
	  *pt=DIRECT;
	  add_to_list(list2,x,y,k,DIRECT);
	}      
      /* PROPAGATE(z,k,INDIRECT) */
      k=Z.k;
      x=Z.x+dx[k];
      y=Z.y+dy[k];
      pt=map+x+sx*y;
      if(inrange(x,y)) if(*pt>INDIRECT)
	{
	  *pt=INDIRECT;
	  add_to_LIST(LIST2,x,y,k);
	}  
    }

    TLIST=LIST1; LIST1=LIST2; LIST2=TLIST;
    empty_LIST(LIST2);

    tlist=list1; list1=list2; list2=tlist;
    empty_list(list2);    
  }

  return(1);
}
