/*

Copyright (c) 2002, Roberto de Alencar Lotufo, and 
Sate University fo Campinas, Brazil.  All rights reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Universidade Estadual
de Campinas not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

THE STATE UNIVERSITY OF CAMPINAS DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL THE STATE UNIVERSITY OF CAMPINAS BE LIABLE
FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

     --------------------------------------------------------------
     lz_edt2 - Lotufo-Zampirolli Exact Euclidean Distance transform
     --------------------------------------------------------------

     This is exact 3-D Euclidean Distance Transform implementation
     based on the paper by
     Lotufo, R. A., and Zampirolli, F. A.
     "Fast Multidimensional Parallel Euclidean Distance Transform Based 
     on Mathematical Morphology.". In: XIV Brazilian Symposium on Computer
     Graphics and Image Processing, 2001, Brazil. IEEE Computer Society, 
     2001, p.100-105. 

     pf    : pointer to input  raster image, unsigned byte, uint8
     pg    : pointer to output raster image, integer, int32
     width : n. of pixels in width
     height: n. of pixels in height
     depth : n. of pixels in depth
*/

#include <stdlib.h>

int lz_edt2( int *pf, int *pg, int w, int h, int d)
{
	int i, j, k, b, y, m, wh;
	unsigned int inf=(0x7fffffff - 2048);
	int *pfj;
	int *pgj, *pgi, *paux;
	
	int *plist1_r, *plist1_l, *plist2_r, *plist2_l;
	int ilist1_r, ilist1_rmax, ilist1_l, ilist1_lmax;
	int ilist2_r, ilist2_l;
	
	wh = w * h;

	/* set all output pixels to infinite,                    */
	for (i=0; i<wh*d; i++) pg[i] = inf;
	
	/* space from queues of the second part of the algorithm */
	plist1_r = (int *) malloc(w*sizeof(int));
	if (plist1_r == NULL) return 0;
	plist2_r = (int *) malloc(w*sizeof(int));
	if (plist2_r == NULL) return 0;
	plist1_l = (int *) malloc(w*sizeof(int));
	if (plist1_l == NULL) return 0;
	plist2_l = (int *) malloc(w*sizeof(int));
	if (plist2_l == NULL) return 0;
	
	for(m=0;m<d;m++) { /* for each slice */
		for(j=0;j<w;j++) { /* for each column j */
			pfj = &pf[j + m*wh];
			pgj = &pg[j + m*wh];
			i=0;
			while (pfj[i*w]) { /* wait for the fist background pixel */
				i++;
				if (i>=h) break;
			}
			if (i<=(h-1)) { /* if is the first time */
				k=i; pgj[k*w]=0;
				b=1; k--;
				while (k >= 0) { /* return up to the beginning of line */
					pgj[k*w] = pgj[(k+1)*w] + b;
					b=b+2;k--;
				}
				i++;
				while (i<=(h-1)) { /* main loop of first pass */
					while (pfj[i*w] == 0) { /* while background pixel, g=0 */
						pgj[i*w]=0; i++;
						if (i >= h) break;
					}
					if (i<=(h-1)) {
						b=1;
						while (pfj[i*w]) { /* while foreground pixel, go up */
							pgj[i*w] = pgj[(i-1)*w] + b;
							b = b + 2; i++;
							if (i >= h) break;
						}
						if (i<=(h-1)) { 
							k=i; pgj[k*w]=0; b=1;
							while (pgj[(k-1)*w] > pgj[k*w]) { /* volta subindo */
								pgj[(k-1)*w] = pgj[k*w] + b;
								b = b + 2; k--;
							}
							i++;
						}
					}
				}
			}
		}
	}
	
	/* second part, erosion by propagation, row by row */
	
	for (m=0; m<d; m++) { /* for each slice */
		for (i=0; i<h; i++) { /* for each column */
			pgi = &pg[i*w + m*wh];
			for (j=0; j<(w-1); j++) {
				plist1_r[j] = w-j-2; /* right propagation, left scanning */
				plist1_l[j] = j+1;   /* left propagation, right scanning */
			}
			ilist1_r    = ilist1_l    = 0;
			ilist1_rmax = ilist1_lmax = w - 1; /* n. of pixels in frontier */
			b = -1;
			while (ilist1_lmax + ilist1_rmax) { /* if any queue not empty */
				b = b + 2;
				ilist2_r = ilist2_l = 0;
				while (ilist1_r < ilist1_rmax) { /* right propagation*/
					j = plist1_r[ ilist1_r]; ilist1_r++; /* remove from queue */
					y = j + 1;
					if (pgi[y] > pgi[j] + b) {
						pgi[y] = pgi[j] + b;
						if (y < (w-1)) {
							plist2_r[ilist2_r] = y; ilist2_r++; /* insert in queue */
						}
					}
				}
				while (ilist1_l < ilist1_lmax) { /* left propagation */
					j = plist1_l[ ilist1_l]; ilist1_l++; /* remove from queue */
					y = j - 1;
					if (pgi[y] > pgi[j] + b) {
						pgi[y] = pgi[j] + b;
						if (y > 0) {
							plist2_l[ilist2_l] = y; ilist2_l++; /* poe na fila */
						}
					}
				}    
				paux = plist1_r; plist1_r = plist2_r; plist2_r = paux;
				paux = plist1_l; plist1_l = plist2_l; plist2_l = paux;
				ilist1_r = ilist1_l = 0;
				ilist1_rmax = ilist2_r;
				ilist1_lmax = ilist2_l;
			}
		}
	}
	/* Third part, erosion by propagation, depth by depth */
	if (d > 1) { /* only if image has more than one slice */
		for (j=0; j<w; j++) { /* for each line */
			for (i=0; i<h; i++) { /* for each column */
				pgi = &pg[j + i*w];
				for (m=0; m<(d-1); m++) {
					plist1_r[m] = d-m-2; /* right propagation, left scanning */
					plist1_l[m] = m+1;   /* left propagation, right scanning */
				}
				ilist1_r    = ilist1_l    = 0;
				ilist1_rmax = ilist1_lmax = d - 1; /* n. of pixels in frontier */
				b = -1;
				while (ilist1_lmax + ilist1_rmax) { /* if any list not empty */
					b = b + 2;
					ilist2_r = ilist2_l = 0;
					while (ilist1_r < ilist1_rmax) { /* right propagation */
						m = plist1_r[ ilist1_r]; ilist1_r++; /* remove from queue */
						y = m + 1;
						if (pgi[y*wh] > pgi[m*wh] + b) {
							pgi[y*wh] = pgi[m*wh] + b;
							if (y < (d-1)) {
								plist2_r[ilist2_r] = y; ilist2_r++; /* insert in queue */
							}
						}
					}
					while (ilist1_l < ilist1_lmax) { /* left  propagation*/
						m = plist1_l[ ilist1_l]; ilist1_l++; /* remove from queue */
						y = m - 1;
						if (pgi[y*wh] > pgi[m*wh] + b) {
							pgi[y*wh] = pgi[m*wh] + b;
							if (y > 0) {
								plist2_l[ilist2_l] = y; ilist2_l++; /* insert in queue */
							}
						}
					}    
					paux = plist1_r; plist1_r = plist2_r; plist2_r = paux;
					paux = plist1_l; plist1_l = plist2_l; plist2_l = paux;
					ilist1_r = ilist1_l = 0;
					ilist1_rmax = ilist2_r;
					ilist1_lmax = ilist2_l;
				}
			}
		}
	}
	free(plist1_r);
	free(plist1_l);
	free(plist2_r);
	free(plist2_l);
	return 1;
}
