/*=======================================================================

  Program:  Distance Transforms
  Module:   $Id: maurer.c,v 1.1.1.1 2002/03/22 10:40:02 cuisenai Exp $
  Date:     $Date: 2002/03/22 10:40:02 $
  Language: C
  
  Author:   Olivier Cuisenaire, http://ltswww.epfl.ch/~cuisenai
            
  Descr.: distmap_maurer(int *map, int n[2]) 

   Takes an image (map) of integers of size n[0]*n[1] as input. It 
   computes for all pixels the square of the Euclidean distance to 
   the nearest zero pixel using Maurer's algorithm presented at 
   IPMI 2001.

  Reference for the algorithm: 
  
    Calvin R. Maurer Jr., Vijay Raghavan, Rensheng Qi, "A Linear Time 
    Algorithm for Computing the Euclidean Distance Transform in Arbitrary 
    Dimensions", IPMI 2001, pp. 358-364
   
  Terms of Use: 

   You can use/modify this program for any use you wish, provided you cite 
   the above references in any publication about it. 
    
  Disclaimer:

   In no event shall the authors or distributors be liable to any party for 
   direct, indirect, special, incidental, or consequential damages arising out 
   of the use of this software, its documentation, or any derivatives thereof, 
   even if the authors have been advised of the possibility of such damage.
   
   The authors and distributors specifically disclaim any warranties, including,
   but not limited to, the implied warranties of merchantability, fitness for a
   particular purpose, and non-infringement.  this software is provided on an 
   "as is" basis, and the authors and distributors have no obligation to provide
   maintenance, support, updates, enhancements, or modifications.
   
=========================================================================*/

#define TRUE 1
#define DMAP(x,y) map[(x)+L*(y)]
#define L maxx
#define M maxy
#define INFINITY 300000000

int distmap_maurer(int *map, int size[2])
{
  int i,j,k;
  int a,b,c,tst,npt,d1,d2;
  int n,m;
  int maxx, maxy;
  int *g,*h,d;
  int *sq;
  int dist;

  maxx=size[0]; maxy=size[1];
  sq=(int*)calloc(2*M+2,sizeof(int));
  for(i=2*M+1;i>=0;i--) sq[i]=i*i;
  
  g=(int *)calloc(M+2,sizeof(int));
  h=(int *)calloc(M+2,sizeof(int));
  
  /* Step1 */
  
  for(j=0;j<M;j++) {
    
    /* forward */
    
    dist=L;
    for(i=0;i<L;i++) {
      if(DMAP(i,j)!=0) dist++;
      else dist=0;
      DMAP(i,j)=sq[dist];
    }
    
    /* backward */
    
    dist=L;
    for(i=L-1;i>=0;i--) {
      if(DMAP(i,j)!=0) dist++;
      else dist=0;
      if(DMAP(i,j)>sq[dist]) DMAP(i,j)=sq[dist];
    }
    
  }
  
  /* step2 */
  
  for(i=0;i<L;i++) {
    
    npt = 0;

    for(j=0;j<M;j++) {

      d = DMAP(i,j);
      if(d!=INFINITY) {
	
	if(npt<2) { 
	  
	   npt++; g[npt] = d; h[npt] = j;
	  
	} else { 
	  
	  a = h[npt] - h[npt-1];
	  b = j - h[npt];
	  c = a+b;
	  tst = c * g[npt] - b * g[npt-1] - a * d - a*b*c ;

	  while( ( npt >=2 ) && ( tst > 0 ) ) { 
	    npt--;
	    a = h[npt] - h[npt-1];
	    b = j - h[npt];
	    c = a+b;
	    tst = c * g[npt] - b * g[npt-1] - a * d - a*b*c ;	    
	  }
	  
	   npt++; g[npt] = d; h[npt] = j;
	}
      }
    }

    k=1;
      
    for(j=0;j<M;j++) {
      d1 = g[k] + (h[k]-j) * (h[k]-j);
      d2 = g[k+1] + (h[k+1]-j) * (h[k+1]-j);
      while ( ( k < npt ) && ( d1 > d2 ) ) { 
	d1 = d2; 
	k++; 
	d2 = g[k+1] + (h[k+1]-j) * (h[k+1]-j);
      }
      DMAP(i,j) = d1;
    }  
  }
 
  return(TRUE);
}

