/*=======================================================================

  Program:  Distance Transforms
  Module:   $Id: saito.c,v 1.1.1.1 2002/03/22 10:40:02 cuisenai Exp $
  Date:     $Date: 2002/03/22 10:40:02 $
  Language: C
  
  Author:   Olivier Cuisenaire, http://ltswww.epfl.ch/~cuisenai
            
  Descr.: distmap_saito2d(int *map, int n[2]) 

   Takes an image (map) of integers of size n[0]*n[1] as input. It 
   computes for all pixels the square of the Euclidean distance to 
   the nearest zero pixel using Saito's algorithm. 

  Reference for the algorithm: 
  
   T. Saito and J.I. Toriwaki, "New algorithms for Euclidean distance 
   transformations of an n-dimensional digitised picture with applications",
   Pattern Recognition, 27(11), pp. 1551-1565, 1994
   
  Reference for the implementation:

   Chapter 2 of "Distance transformations: fast algorithms and applications 
   to medical image processing", Olivier Cuisenaire's Ph.D. Thesis, October 
   1999, Universit catholique de Louvain, Belgium.
    
  Terms of Use: 

   You can use/modify this program for any use you wish, provided you cite 
   the above references in any publication about it. 
    
  Disclaimer:

   In no event shall the authors or distributors be liable to any party for 
   direct, indirect, special, incidental, or consequential damages arising out 
   of the use of this software, its documentation, or any derivatives thereof, 
   even if the authors have been advised of the possibility of such damage.
   
   The authors and distributors specifically disclaim any warranties, including,
   but not limited to, the implied warranties of merchantability, fitness for a
   particular purpose, and non-infringement.  this software is provided on an 
   "as is" basis, and the authors and distributors have no obligation to provide
   maintenance, support, updates, enhancements, or modifications.
   
=========================================================================*/

#include "distmaps.h"

int distmap_saito2d(int *map, int size[2])
{

  int i,j,k;
  int a,b,*df2;
  int n,m;
  int N,M,L;
  int *npt, *pt, *ptstart,*ptend;
  int *buff,buffer;
  int *sq;

  N=1;M=size[1]; L=size[0];
  sq=(int*)calloc(2*M+2,sizeof(int));
  for(i=2*M+1;i>=0;i--) sq[i]=i*i;

  buff=(int *)calloc(M,sizeof(int));

  for(k=0;k<N;k++) { 
    /* for all slices */

    /* Step1 */

    for(j=0;j<M;j++) {
      
      ptstart=map+L*j;
      ptend=ptstart+L;
      
      /* forward */
      
      df2=sq+M;
      for(pt=ptstart;pt<ptend;pt++) {
	if(*pt!=0) df2++;
	else df2=sq;
	*pt=*df2;
      }
    
      /* backward */
   
      df2=sq+M;
      for(pt=ptend-1;pt>=ptstart;pt--) {
	if(*pt!=0) df2++;
	else df2=sq;
	if(*pt>*df2) *pt=*df2;
      }

    }

    /* Step2 */   

    for(i=0;i<L;i++) {
      
      pt=map+i;
      
      for(j=0;j<M;j++,pt+=L) {
        buff[j]=*pt;
      }
      

      pt=map+i+L;
      a=0; buffer=buff[0];
      
      for(j=1;
	  j<M;
	  j++,pt+=L) 
	{
	 	  
	  if(a>0) a--;
	  if(buff[j]>buffer+1) {
	    b=(buff[j]-buffer-1)/2;
	    if((j+b)>M-1) b=(M-1)-j;
	    
	    npt=pt+a*M;
	    for(n=a;n<=b;n++) {
	      m=buffer+sq[n+1];
	      if(buff[j+n]<=m) n=b;
	      else if(m<*npt) *npt=m;
	      npt+=M;
	    }
	    a=b; 
	  }
	  else
	    a=0;
	  buffer=buff[j];
	}
      
      
      a=0;
      pt-=2*L;
      buffer=buff[M-1];
      
      for(j=M-2;
	  j>=0;
	  j--,pt-=L)
	{
	  if(a>0) a--;
	  if(buff[j]>buffer+1)
	    {
	      b=(buff[j]-buffer-1)/2;
	      if((j-b)<0) b=j;
	      
	      npt=pt-a*M;
	      for(n=a;n<=b;n++) 
		{
		  m=buffer+sq[n+1];
		  if(buff[j-n]<=m) 
		     n=b;
		  else 
		    if(m<*npt) *npt=m;
		  npt-=M;
		}
	      a=b;
	    }
	  else
	    a=0;
	  buffer=buff[j];
	}
      
      
    }
    
  }

  return(1);
}

