/*=======================================================================

  Program:  Distance Transforms
  Module:   $Id: saito_yx.c,v 1.1.1.1 2002/03/22 10:40:02 cuisenai Exp $
  Date:     $Date: 2002/03/22 10:40:02 $
  Language: C
  
  Author:   Olivier Cuisenaire, http://ltswww.epfl.ch/~cuisenai
            
  Descr.: distmap_saito2d_yx(int *map, int n[2]) 

   Takes an image (map) of integers of size n[0]*n[1] as input. It 
   computes for all pixels the square of the Euclidean distance to 
   the nearest zero pixel using Saito's algorithm. This version works 
   first along the columns, then along the lines, which is better for
   caching of the heavier second scan. 

  Reference for the algorithm: 
  
   T. Saito and J.I. Toriwaki, "New algorithms for Euclidean distance 
   transformations of an n-dimensional digitised picture with applications",
   Pattern Recognition, 27(11), pp. 1551-1565, 1994
      
  Reference for the implementation:

   Chapter 2 of "Distance transformations: fast algorithms and applications 
   to medical image processing", Olivier Cuisenaire's Ph.D. Thesis, October 
   1999, Universit catholique de Louvain, Belgium.
    
  Terms of Use: 

   You can use/modify this program for any use you wish, provided you cite 
   the above references in any publication that uses it. 
    
  Disclaimer:

   In no event shall the authors or distributors be liable to any party for 
   direct, indirect, special, incidental, or consequential damages arising out 
   of the use of this software, its documentation, or any derivatives thereof, 
   even if the authors have been advised of the possibility of such damage.
   
   The authors and distributors specifically disclaim any warranties, including,
   but not limited to, the implied warranties of merchantability, fitness for a
   particular purpose, and non-infringement.  this software is provided on an 
   "as is" basis, and the authors and distributors have no obligation to provide
   maintenance, support, updates, enhancements, or modifications.
   
=========================================================================*/
 
#define TRUE 1
#define DMAP(x,y) map[(y)+L*(x)]
#define L maxy
#define M maxx

int distmap_saito2d_yx(int *map, int size[2])
{
  int i,j;
  int a,b;
  int n,m;
  int maxx, maxy;
  int *buff,buffer;
  int *sq;
  int dist;

  maxx=size[0]; maxy=size[1];
  sq=(int*)calloc(2*M+2,sizeof(int));
  for(i=2*M+1;i>=0;i--) sq[i]=i*i;
  
  buff=(int *)calloc(M,sizeof(int));
  
  /* Step1 */
  
  for(j=0;j<M;j++) {
    
    /* forward */
    
    dist=L;
    for(i=0;i<L;i++) {
      if(DMAP(i,j)!=0) dist++;
      else dist=0;
      DMAP(i,j)=sq[dist];
    }
    
    /* backward */
    
    dist=L;
    for(i=L-1;i>=0;i--) {
      if(DMAP(i,j)!=0) dist++;
      else dist=0;
      if(DMAP(i,j)>sq[dist]) DMAP(i,j)=sq[dist];
    }
    
  }
  
  /* Step2 */   
  
  for(i=0;i<L;i++) {
    
    for(j=0;j<M;j++) {
      buff[j]=DMAP(i,j);
    }
    
    a=0; buffer=buff[0];
    
    for(j=1;
	j<M;
	j++) 
      {
	
	if(a>0) a--;
	if(buff[j]>buffer+1) {
	  b=(buff[j]-buffer-1)/2;
	  if((j+b)>M-1) b=(M-1)-j;
	  
	  for(n=a;n<=b;n++) {
	    m=buffer+sq[n+1];
	    if(buff[j+n]<=m) n=b;
	    else if(m<DMAP(i,j+n)) DMAP(i,j+n)=m;
	  }
	  a=b; 
	}
	else
	  a=0;
	buffer=buff[j];
      }
    
    
    a=0;
    buffer=buff[M-1];
    
    for(j=M-2;
	j>=0;
	j--)
      {
	if(a>0) a--;
	if(buff[j]>buffer+1)
	  {
	    b=(buff[j]-buffer-1)/2;
	    if((j-b)<0) b=j;
	    
	    for(n=a;n<=b;n++) 
	      {
		m=buffer+sq[n+1];
		if(buff[j-n]<=m) 
		  n=b;
		else 
		  if(m<DMAP(i,j-n)) DMAP(i,j-n)=m;
	      }
	    a=b;
	  }
	else
	  a=0;
	buffer=buff[j];
      }
  }

  return(TRUE);
}

